package org.jvnet.hifaces.application.navigation;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.faces.FacesException;
import javax.faces.application.NavigationHandler;
import javax.faces.application.ViewHandler;
import javax.faces.component.UIViewRoot;
import javax.faces.context.ExternalContext;
import javax.faces.context.FacesContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.myfaces.config.RuntimeConfig;
import org.apache.myfaces.config.element.NavigationCase;
import org.apache.myfaces.config.element.NavigationRule;
import org.apache.myfaces.portlet.PortletUtil;
import org.apache.myfaces.shared.util.HashMapUtils;

public class DelegatingNavigationHandler extends NavigationHandler {

	private static final Log log = LogFactory
			.getLog(DelegatingNavigationHandler.class);

	private static final String ASTERISK = "*";

	private Map _navigationCases = null;

	private List _wildcardKeys = new ArrayList();

	private NavigationHandler original;

	public DelegatingNavigationHandler(NavigationHandler original) {
		this.original = original;
	}

	public void handleNavigation(FacesContext facesContext, String fromAction,
			String outcome) {
		if (outcome == null) {
			// stay on current ViewRoot
			return;
		}

		String viewId = facesContext.getViewRoot().getViewId();
		Map casesMap = getNavigationCases(facesContext);
		NavigationCase navigationCase = null;

		List casesList = (List) casesMap.get(viewId);
		if (casesList != null) {
			// Exact match?
			navigationCase = calcMatchingNavigationCase(casesList, fromAction,
					outcome);
		}

		if (navigationCase == null) {
			// Wildcard match?
			List keys = getSortedWildcardKeys();
			for (int i = 0, size = keys.size(); i < size; i++) {
				String fromViewId = (String) keys.get(i);
				if (fromViewId.length() > 2) {
					String prefix = fromViewId.substring(0,
							fromViewId.length() - 1);
					if (viewId != null && viewId.startsWith(prefix)) {
						casesList = (List) casesMap.get(fromViewId);
						if (casesList != null) {
							navigationCase = calcMatchingNavigationCase(
									casesList, fromAction, outcome);
							if (navigationCase != null)
								break;
						}
					}
				} else {
					casesList = (List) casesMap.get(fromViewId);
					if (casesList != null) {
						navigationCase = calcMatchingNavigationCase(casesList,
								fromAction, outcome);
						if (navigationCase != null)
							break;
					}
				}
			}
		}

		final Navigator navigator = getNavigator(facesContext);
		if (navigationCase != null) {
			if (log.isTraceEnabled()) {
				log.trace("handleNavigation fromAction=" + fromAction
						+ " outcome=" + outcome + " toViewId ="
						+ navigationCase.getToViewId() + " redirect="
						+ navigationCase.isRedirect());
			}
			if (navigationCase.isRedirect()
					&& (!PortletUtil.isPortletRequest(facesContext))) { // Spec
				if (navigator == null) {
					redirect(facesContext, navigationCase.getToViewId());
				} else {
					navigator.navigate(facesContext, fromAction, outcome,
							navigationCase.getToViewId(), true);
				}
			} else {
				if (navigator == null) {
					forward(facesContext, navigationCase.getToViewId());
				} else {
					navigator.navigate(facesContext, fromAction, outcome,
							navigationCase.getToViewId(), false);
				}

			}
		} else {
			if (navigator == null) {
				// no navigationcase found, stay on current ViewRoot
				if (log.isTraceEnabled()) {
					log
							.trace("handleNavigation fromAction="
									+ fromAction
									+ " outcome="
									+ outcome
									+ " no matching navigation-case found, staying on current ViewRoot");
				}
			} else {
				navigator.reload(facesContext, fromAction, outcome);
			}
		}
	}

	protected void forward(FacesContext facesContext, String toViewId) {
		ViewHandler viewHandler = facesContext.getApplication()
				.getViewHandler();
		// create new view
		UIViewRoot viewRoot = viewHandler.createView(facesContext, toViewId);
		facesContext.setViewRoot(viewRoot);
		facesContext.renderResponse();
	}

	protected void redirect(FacesContext facesContext, String toViewId) {
		ExternalContext externalContext = facesContext.getExternalContext();
		ViewHandler viewHandler = facesContext.getApplication()
				.getViewHandler();
		String redirectPath = viewHandler.getActionURL(facesContext, toViewId);

		try {
			externalContext.redirect(externalContext
					.encodeActionURL(redirectPath));
		} catch (IOException e) {
			throw new FacesException(e.getMessage(), e);
		}
	}

	private NavigationCase calcMatchingNavigationCase(List casesList,
			String actionRef, String outcome) {
		for (int i = 0, size = casesList.size(); i < size; i++) {
			NavigationCase caze = (NavigationCase) casesList.get(i);
			String cazeOutcome = caze.getFromOutcome();
			String cazeActionRef = caze.getFromAction();
			if ((cazeOutcome == null || cazeOutcome.equals(outcome))
					&& (cazeActionRef == null || cazeActionRef
							.equals(actionRef))) {
				return caze;
			}
		}
		return null;
	}

	private List getSortedWildcardKeys() {
		return _wildcardKeys;
	}

	private Map getNavigationCases(FacesContext facesContext) {
		ExternalContext externalContext = facesContext.getExternalContext();
		RuntimeConfig runtimeConfig = RuntimeConfig
				.getCurrentInstance(externalContext);

		if (_navigationCases == null
				|| runtimeConfig.isNavigationRulesChanged()) {
			synchronized (this) {
				if (_navigationCases == null
						|| runtimeConfig.isNavigationRulesChanged()) {
					Collection rules = runtimeConfig.getNavigationRules();
					int rulesSize = rules.size();
					Map cases = new HashMap(HashMapUtils
							.calcCapacity(rulesSize));
					List wildcardKeys = new ArrayList();

					for (Iterator iterator = rules.iterator(); iterator
							.hasNext();) {
						NavigationRule rule = (NavigationRule) iterator.next();
						String fromViewId = rule.getFromViewId();

						// specification 7.4.2 footnote 4 - missing fromViewId
						// is allowed:
						if (fromViewId == null) {
							fromViewId = ASTERISK;
						} else {
							fromViewId = fromViewId.trim();
						}

						List list = (List) cases.get(fromViewId);
						if (list == null) {
							list = new ArrayList(rule.getNavigationCases());
							cases.put(fromViewId, list);
							if (fromViewId.endsWith(ASTERISK)) {
								wildcardKeys.add(fromViewId);
							}
						} else {
							list.addAll(rule.getNavigationCases());
						}

					}
					Collections.sort(wildcardKeys, new KeyComparator());

					synchronized (cases) {
						// We do not really need this sychronization at all, but
						// this
						// gives us the peace of mind that some good optimizing
						// compiler
						// will not rearrange the execution of the assignment to
						// an
						// earlier time, before all init code completes
						_navigationCases = cases;
						_wildcardKeys = wildcardKeys;

						runtimeConfig.setNavigationRulesChanged(false);
					}
				}
			}
		}
		return _navigationCases;
	}

	private static final class KeyComparator implements Comparator {
		public int compare(Object o1, Object o2) {
			return -(((String) o1).compareTo((String) o2));
		}
	}

	public static final String NAVIGATOR_BEAN_NAME = "navigator";

	public Navigator getNavigator(FacesContext facesContext) {
		return (Navigator) facesContext.getApplication().getVariableResolver()
				.resolveVariable(facesContext, NAVIGATOR_BEAN_NAME);
	}
}
